/* 
 * Copyright (c) 2007-2008
 *     IS/AIST-ST2I/CNRS Joint Japanese-French Robotics Laboratory (JRL).
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include <cxtypes.h>
# include <cv.h>
#endif


/* --- local data ---------------------------------------------------- */

struct viam_camera_t {
  struct viam_camera_t *next;

  char *	name;		/**< local name */
  viam_hwdev_t *hwdev;		/**< hardware device */
  char *	uid;		/**< unique id */

  int		setup;		/**< true if camera is configured */

  viam_hwmode_t hwmode;		/**< hw acquisition mode */

  IplImage	raw;		/**< header of raw images */

  void *	hw_data;	/**< hardware specific data */
  void *	usr_data;	/**<\brief user data */
};


/* --- viam_camera_create -------------------------------------------- */

/** \brief Create a new camera in the module configuration
 *
 * \param[in] vh	Library handle
 * \param[in] id	Unique camera ID of the form dev:uid
 * \param[in] name	Local name
 *
 * \return camera object
 */

viam_camera_t
viam_camera_create(viam_handle_t vh, const char *id, const char *name)
{
  char *uid;
  viam_hwdev_t *hwdev;
  viam_camera_t c;

  hwdev = viam_hardware_device(vh, id, &uid);
  if (!uid) return NULL;

  if (viam_camera_getbyuid(vh, uid) || viam_camera_getbyname(vh, name)) {
    warnx("camera `%s' already exists", name);
    viam_seterrno(vh, Eviamlib_DUPLICATE_ID);
    return NULL;
  }

  c = malloc(sizeof(*c));
  if (!c) {
    warnx("unable to allocate memory for camera %s", name);
    viam_seterrno(vh, Eviamlib_MEMORY);
    return NULL;
  }

  c->name = strdup(name);
  c->hwdev = hwdev;
  c->uid = strdup(uid);
  c->setup = 0;
  c->hwmode.size = VIAM_HWSZ_INVALID;
  c->hwmode.format = VIAM_HWFMT_INVALID;
  c->hwmode.crop = VIAM_HW_FIXED;
  c->hwmode.fps = VIAM_HWFPS_INVALID;
  c->hw_data = NULL;
  c->usr_data = NULL;

  c->next = viam_config_cameras(vh);
  viam_config_setcameras(vh, c);
  return c;
}


/* --- viam_camera_delete -------------------------------------------- */

/** \brief Destroy a camera
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 */

void
viam_camera_delete(viam_handle_t vh, viam_camera_t camera)
{
  viam_camera_t c;

  if (camera == viam_config_cameras(vh)) {
    viam_config_setcameras(vh, camera->next);
  } else for(c = viam_config_cameras(vh); c->next; c = c->next)
    if (camera == c->next) {
      c->next = camera->next;
    }

  if (camera->name) free(camera->name);
  if (camera->uid) free(camera->uid);
  free(camera);
}


/* --- viam_camera_name ---------------------------------------------- */

/** \brief Return local name of the camera
 *
 * \param[in] cam	Camera handle
 *
 * \return camera name
 */

const char *
viam_camera_name(viam_camera_t cam)
{
  return cam->name;
}


/* --- viam_camera_uid ----------------------------------------------- */

/** \brief Return camera unique identifier
 *
 * \param[in] c		Camera handle
 *
 * \return camera name
 */

const char *
viam_camera_uid(viam_camera_t c)
{
  return c->uid;
}


/* --- viam_camera_getbyuid ------------------------------------------ */

/** \brief Get a camera by its uid
 *
 * \param[in] vh	Library handle.
 * \param[in] uid	Unique camera ID on the 1394 bus
 *
 * \return the camera handle, or NULL if not found
 */

viam_camera_t
viam_camera_getbyuid(const viam_handle_t vh, const char *uid)
{
  viam_camera_t c;

  for(c = viam_config_cameras(vh); c; c = c->next)
    if (!strcmp(c->uid, uid)) return c;

  return NULL;
}


/* --- viam_camera_getbyname ----------------------------------------- */

/** \brief Get a camera by its local name
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Local name of the camera.
 *
 * \return the camera handle, or NULL if not found
 */

viam_camera_t
viam_camera_getbyname(const viam_handle_t vh, const char *name)
{
  viam_camera_t c;

  for(c = viam_config_cameras(vh); c; c = c->next)
    if (!strcmp(c->name, name)) return c;

  return NULL;
}


/* --- viam_camera mutators and accessors ---------------------------- */

viam_camera_t
viam_camera_next(viam_camera_t camera)
{
  return camera->next;
}

viam_hwdev_t *
viam_camera_hwdev(viam_camera_t camera)
{
  return camera->hwdev;
}

void
viam_camera_sethwdev(viam_camera_t camera, viam_hwdev_t *hwdev)
{
  camera->hwdev = hwdev;
}

int
viam_camera_setup(viam_camera_t camera)
{
  return camera->setup;
}

void
viam_camera_setsetup(viam_camera_t camera, int setup)
{
  camera->setup = setup;
}

const viam_hwmode_t *
viam_camera_hwmode(viam_handle_t vh, viam_camera_t camera)
{
  if (camera->hwmode.size == VIAM_HWSZ_INVALID ||
      camera->hwmode.format == VIAM_HWFMT_INVALID ||
      camera->hwmode.fps == VIAM_HWFPS_INVALID) {
    viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
    return NULL;
  }

  return &camera->hwmode;
}

int
viam_camera_sethwmode(viam_handle_t vh, viam_camera_t camera,
		      const viam_hwmode_t *hwmode)
{
  CvSize size;
  IplImage *raw;
  int depth, channels;
  viam_image_t im;
  int s;

  if (viam_camera_setup(camera))
    return viam_seterrno(vh, Eviamlib_CONFIGURED);

  /* check mode is valid and get size */
  s = viam_hwmode_size(vh, hwmode,
		       &size.width, &size.height, &depth, &channels);
  if (s) return s;

  raw = viam_camera_raw(camera);
  cvInitImageHeader(raw, size, depth, channels, IPL_ORIGIN_TL, 4);
  viam_setswap(raw->imageSize);

  warnx("camera %s produces %dx%d images, %d bpp, %d channel%s",
	viam_camera_name(camera), size.width, size.height,
	depth, channels, channels>1?"s":"");

  /* reconfigure filters on images that use this camera */
  for (im = viam_config_images(vh); im; im = viam_image_next(im))
    if (viam_image_camera(im) == camera) {
      s = viam_filter_configure(vh, im, viam_image_flist(im));
      if (s) {
	warnx("could not set mode for camera %s", viam_camera_name(camera));
	return s;
      }
    }

  camera->hwmode = *hwmode;

  return 0;
}

IplImage *
viam_camera_raw(viam_camera_t camera)
{
  return &camera->raw;
}

void *
viam_camera_hwdata(viam_camera_t camera)
{
  return camera->hw_data;
}

void
viam_camera_sethwdata(viam_camera_t camera, void *hwdata)
{
  camera->hw_data = hwdata;
}

void *
viam_camera_usrdata(viam_camera_t camera)
{
  return camera->usr_data;
}

void
viam_camera_setusrdata(viam_camera_t camera, void *usrdata)
{
  camera->usr_data = usrdata;
}
